﻿using System;
using System.Collections.Generic;
using System.IO;
using System.IO.Compression;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Xml.Linq;

namespace ExcelReader
{
    /// <summary>
    /// セル（行、桁、テキストを保持）
    /// </summary>
    public class Cell
    {
        public int Row { get; set; }
        public int Column { get; set; }
        public string Text { get; set; }
    }

    /// <summary>
    /// ワークシート
    /// </summary>
    public class Sheet
    {
        private string _name = string.Empty;
        private int _rowsize = 0;
        private int _columnsize = 0;
        private List<Cell> _cells = new List<Cell>();

        public Sheet(string aName)
        {
            _name = aName;
        }

        // ワークシートの名前
        public string Name { get { return _name; } }
        // データのある最大行
        public int RowSize { get { return _rowsize + 1; } }
        // データのある最大桁
        public int ColumnSize { get { return _columnsize + 1; } }
        // データのあるセルのリスト
        public List<Cell> Cells { get { return _cells; } }
        // 指定された行/桁のデータ（すべて文字列として保存されている）
        public string this[int r, int c]
        {
            get
            {
                var cell = _cells.Where(x => x.Row == r && x.Column == c).SingleOrDefault();
                return cell == null ? null : cell.Text;
            }
            set
            {
                var cell = _cells.Where(x => x.Row == r && x.Column == c).SingleOrDefault();
                if (cell == null)
                {
                    _cells.Add(new Cell() { Row = r, Column = c, Text = value });
                    if (_rowsize < r)
                        _rowsize = r;
                    if (_columnsize < c)
                        _columnsize = c;
                }
                else
                {
                    cell.Text = value;
                }
            }
        }
    }

    /// <summary>
    /// ワークブック
    /// </summary>
    public class Workbook
    {
        private List<Sheet> _sheets = new List<Sheet>();

        public Sheet this[int index]
        {
            get
            {
                if (index < 0 || _sheets.Count <= index)
                    return null;
                return _sheets[index];
            }
        }

        public Sheet this[string name]
        {
            get
            {
                return _sheets.Where(x => x.Name == name).SingleOrDefault();
            }
        }

        public List<Sheet> Sheets { get { return _sheets; } }
        public int SheetCount { get { return _sheets.Count; } }
    }

    /// <summary>
    /// Excel 読み込みクラス（staticクラス）
    /// </summary>
    public static class ExcelReader
    {
        public static Workbook Load(string excelfile)
        {
            return Load(new FileStream(excelfile, FileMode.Open));
        }

        public static Workbook Load(Stream zipstream)
        {
            using (var zip = new ZipArchive(zipstream))
            {
                var wbxml = zip.Entries.Where(x => x.FullName == "xl/workbook.xml").SingleOrDefault();
                // workbook が空であってはならない
                if (wbxml == null)
                    return null;

                // 共有文字列があれば、リストとして取り込んでおく
                var slist = new List<string>();
                var ssxml = zip.Entries.Where(x => x.FullName == "xl/sharedStrings.xml").SingleOrDefault();
                if (ssxml != null)
                {
                    var ssdoc = XDocument.Load(ssxml.Open());
                    foreach (var node in ssdoc.Descendants().Where(x => x.Name.LocalName == "si"))
                    {
                        if (node.FirstNode is XElement)
                            slist.Add(((XElement)node.FirstNode).Value);
                    }
                }

                var workbook = new Workbook();
                var xdoc = XDocument.Load(wbxml.Open());
                var xsheet = xdoc.Descendants().Where(x => x.Name.LocalName == "sheet");
                var id = 1;
                foreach (var xs in xsheet)
                {
                    // var id = xs.Attribute("sheetId").Value;  // 違った
                    var name = xs.Attribute("name").Value;
                    var sxml = zip.Entries.Where(x => x.FullName == string.Format("xl/worksheets/sheet{0}.xml", id)).SingleOrDefault();
                    id++;
                    if (sxml == null)
                        continue;

                    var sdoc = XDocument.Load(sxml.Open());
                    var sheet = new Sheet(name);
                    foreach (var row in sdoc.Descendants().Where(x => x.Name.LocalName == "row"))
                    {
                        for (var item = row.FirstNode as XElement; item != null; item = item.NextNode as XElement)
                        {
                            if (item.Name.LocalName == "c")
                            {
                                var x = item.Attribute("r");
                                if (x != null)
                                {
                                    var str = x.Value;  // str は "A1" のようなセル位置をあらわす文字列
                                    int c = (int)(str[0] - 'A');        // bugfix: 2016/1/16
                                    int i = 1;
                                    while ('A' <= str[i] && str[i] <= 'Z')      
                                    {
                                        c = (c + 1) * 26 + (int)(str[i] - 'A');
                                        i++;
                                    }
                                    int r = 0;
                                    while (i < str.Length)
                                    {
                                        r = r * 10 + (int)(str[i] - '0');
                                        i++;
                                    }
                                    r--;
                                    // r, c でセル位置（0起点の行、桁）

                                    var vitem = item.FirstNode as XElement;
                                    while (vitem != null && vitem.Name.LocalName != "v")
                                        vitem = vitem.NextNode as XElement;
                                    if (vitem == null)
                                        continue;

                                    var t = item.Attribute("t");    // 属性を取得
                                    if (t == null)
                                    {
                                        // t 属性がないものは、おそらく数値
                                        // 形式は s 属性で指定されるようだが、詳細は不明（s.Value が何かのインデックス）
                                        var s = item.Attribute("s");
                                        // 文字列のままセルの値として設定
                                        sheet[r, c] = vitem.Value;
                                        //    // 日付の場合
                                        //    sheet[r, c] = string.IsNullOrEmpty(vitem.Value) ? string.Empty : DateTime.FromOADate(Convert.ToDouble(vitem.Value)).ToString();
                                        //    // TimeSpan の場合
                                        //    sheet[r, c] = string.IsNullOrEmpty(vitem.Value) ? string.Empty : TimeSpan.FromDays(Convert.ToDouble(vitem.Value)).ToString();
                                    }
                                    else if (t.Value == "s")
                                    {
                                        // t 属性が "s" のものは、おそらく文字列（共有文字列のインデックス）
                                        sheet[r, c] = slist[Convert.ToInt32(vitem.Value)];
                                    }
                                    else if (t.Value == "b")
                                    {
                                        // t 属性が "b" ならば論理型
                                        sheet[r, c] = vitem.Value == "1" ? "TRUE" : "FALSE";
                                    }
                                    else
                                    {
                                        // 不明のものは、そのまま設定する
                                        sheet[r, c] = vitem.Value;
                                    }
                                }
                            }
                        }
                    }
                    workbook.Sheets.Add(sheet);
                }
                return workbook;
            }
        }
    }
}
